@@ -95,7 +95,13 @@
     ?S1GW_CTR_S1AP_PROXY_OUT_PKT_FWD_PROC,                  %% forwarded: processed
     ?S1GW_CTR_S1AP_PROXY_OUT_PKT_FWD_UNMODIFIED,            %% forwarded: unmodified
     ?S1GW_CTR_S1AP_PROXY_OUT_PKT_REPLY_ALL,                 %% replied: total
-    ?S1GW_CTR_S1AP_PROXY_OUT_PKT_REPLY_ERAB_SETUP_RSP       %% replied: E-RAB SETUP.rsp
+    ?S1GW_CTR_S1AP_PROXY_OUT_PKT_REPLY_ERAB_SETUP_RSP,      %% replied: E-RAB SETUP.rsp
+
+    %% SCTP related counters
+    ?S1GW_CTR_SCTP_ERROR_ALL,                               %% total number of SCTP errors
+    ?S1GW_CTR_SCTP_ERROR_SEND_FAILED,                       %% send operation failed
+    ?S1GW_CTR_SCTP_ERROR_PDAPI_EVENT,                       %% partial delivery failure
+    ?S1GW_CTR_SCTP_ERROR_REMOTE_ERROR                       %% remote error
 ]).
 
 -define(S1GW_GAUGES, [
